//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


// SearchDlg.cpp : implementation file
//

#include "stdafx.h"
#include "emule.h"
#include "SearchDlg.h"
#include "packets.h"
#include "server.h"
#include "opcodes.h"
#include "otherfunctions.h"
#include "ED2KLink.h"
#include "SearchList.h"
#include "CustomAutoComplete.h"
#include "SearchExpr.h"
#define USE_FLEX
#include "Parser.hpp"
#include "Scanner.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#define	SEARCH_STRINGS_PROFILE	_T("AC_SearchStrings.dat")

extern int yyparse();
extern int yyerror(const char* errstr);

// CSearchDlg dialog

IMPLEMENT_DYNAMIC(CSearchDlg, CDialog)
CSearchDlg::CSearchDlg(CWnd* pParent /*=NULL*/)
	: CResizableDialog(CSearchDlg::IDD, pParent){
	m_nSearchID = 0x80000000;
	m_pacSearchString = NULL;
}

CSearchDlg::~CSearchDlg(){
	if (globsearch) delete searchpacket;
	DestroyIcon(icon_search);
	if (m_pacSearchString){
		m_pacSearchString->Unbind();
		m_pacSearchString->Release();
	}
	
}

BOOL CSearchDlg::OnInitDialog(){
	CResizableDialog::OnInitDialog();
	theApp.searchlist->SetOutputWnd(&searchlistctrl);
	searchlistctrl.Init(theApp.searchlist);
	Localize();
	searchprogress.SetStep(1);
	global_search_timer = 0;
	globsearch = false;
	m_guardCBPrompt=false;

	m_ctrlSearchFrm.Init(IDI_NORMALSEARCH);
	m_ctrlDirectDlFrm.Init(IDI_DIRECTDOWNLOAD);

	icon_search=(HICON)::LoadImage(AfxGetInstanceHandle(), MAKEINTRESOURCE(IDI_SEARCHRESULTS), IMAGE_ICON, 16, 16, 0);
	((CStatic*)GetDlgItem(IDC_SEARCHLST_ICO))->SetIcon(icon_search);
	
	m_ImageList.Create(16,16,theApp.m_iDfltImageListColorFlags|ILC_MASK,0,10);
	m_ImageList.SetBkColor(CLR_NONE);
	m_ImageList.Add(theApp.LoadIcon(IDI_BN_SEARCH));
	searchselect.SetImageList(&m_ImageList);

	AddAnchor(IDC_SDOWNLOAD,BOTTOM_LEFT);
	AddAnchor(IDC_SEARCHLIST,TOP_LEFT,BOTTOM_RIGHT);
	AddAnchor(IDC_PROGRESS1,BOTTOM_LEFT,BOTTOM_RIGHT);

	AddAnchor(IDC_DDOWN_FRM, TOP_LEFT, TOP_RIGHT);
	AddAnchor(IDC_ELINK, TOP_LEFT, TOP_RIGHT);
	AddAnchor(IDC_STARTS, TOP_RIGHT);
	AddAnchor(IDC_CANCELS, TOP_RIGHT);
	AddAnchor(IDC_CLEARALL, TOP_RIGHT);
	AddAnchor(IDC_SEARCH_RESET, TOP_LEFT);
	AddAnchor(searchselect.m_hWnd,TOP_LEFT,TOP_RIGHT);
	AddAnchor(IDC_STATIC_DLTOof,BOTTOM_LEFT);
	AddAnchor(IDC_CATTAB2,BOTTOM_LEFT);
	ShowSearchSelector(false);
	m_lastclpbrd="";

	if (theApp.glob_prefs->GetUseAutocompletion()){
		m_pacSearchString = new CCustomAutoComplete();
		m_pacSearchString->AddRef();
		if (m_pacSearchString->Bind(::GetDlgItem(m_hWnd, IDC_SEARCHNAME), ACO_UPDOWNKEYDROPSLIST | ACO_AUTOSUGGEST))
			m_pacSearchString->LoadList(CString(theApp.glob_prefs->GetConfigDir()) +  _T("\\") SEARCH_STRINGS_PROFILE);
	}

	ASSERT( (GetDlgItem(IDC_EDITSEARCHMIN)->GetStyle() & ES_NUMBER) == 0 );
	ASSERT( (GetDlgItem(IDC_EDITSEARCHMAX)->GetStyle() & ES_NUMBER) == 0 );
	((CEdit*)GetDlgItem(IDC_SEARCHNAME))->LimitText(512); // max. length of search expression
	((CEdit*)GetDlgItem(IDC_EDITSEARCHEXTENSION))->LimitText(8); // max. length of file (type) extension

	return true;
}

void CSearchDlg::DoDataExchange(CDataExchange* pDX){
	CResizableDialog::DoDataExchange(pDX);
	DDX_Control(pDX, IDC_SEARCHLIST, searchlistctrl);
	DDX_Control(pDX, IDC_PROGRESS1, searchprogress);
	DDX_Control(pDX, IDC_COMBO1, methodBox);
	DDX_Control(pDX, IDC_TypeSearch, Stypebox);
	DDX_Control(pDX, IDC_TAB1, searchselect);
	DDX_Control(pDX, IDC_SEARCH_FRM, m_ctrlSearchFrm);
	DDX_Control(pDX, IDC_DDOWN_FRM, m_ctrlDirectDlFrm);
	DDX_Control(pDX, IDC_CATTAB2, m_cattabs);
}


BEGIN_MESSAGE_MAP(CSearchDlg, CResizableDialog)
	ON_BN_CLICKED(IDC_STARTS, OnBnClickedStarts)
	ON_WM_TIMER()
	ON_BN_CLICKED(IDC_CANCELS, OnBnClickedCancels)
	ON_BN_CLICKED(IDC_SDOWNLOAD, OnBnClickedSdownload)
	ON_NOTIFY(NM_DBLCLK, IDC_SEARCHLIST, OnNMDblclkSearchlist)
	ON_BN_CLICKED(IDC_CLEARALL, OnBnClickedClearall)
	ON_NOTIFY(TCN_SELCHANGE, IDC_TAB1, OnTcnSelchangeTab1)
	ON_EN_CHANGE(IDC_SEARCHNAME, OnEnChangeSearchname)
	ON_CBN_SELCHANGE(IDC_TypeSearch, OnEnChangeSearchname)
	ON_MESSAGE(WM_CLOSETAB, OnCloseTab)
	ON_EN_CHANGE(IDC_SEARCHAVAIL , OnEnChangeSearchname)
	ON_EN_CHANGE(IDC_SEARCHEXTENTION, OnEnChangeSearchname)
	ON_EN_CHANGE(IDC_SEARCHMINSIZE, OnEnChangeSearchname)
	ON_EN_CHANGE(IDC_SEARCHMAXSIZE, OnEnChangeSearchname)
	ON_BN_CLICKED(IDC_SEARCH_RESET, OnBnClickedSearchReset)
	ON_EN_KILLFOCUS(IDC_ELINK, OnEnKillfocusElink)
END_MESSAGE_MAP()


// CSearchDlg message handlers

void CSearchDlg::OnBnClickedStarts(){
	searchprogress.SetPos(0);

	// ed2k-links
	if (GetDlgItem(IDC_ELINK)->GetWindowTextLength()){
		CString strlink;
		GetDlgItem(IDC_ELINK)->GetWindowText(strlink);
		GetDlgItem(IDC_ELINK)->SetWindowText("");

		AddEd2kLinksToDownload(strlink);
		return;
	}

	// start "normal" server-search
	if (GetDlgItem(IDC_SEARCHNAME)->GetWindowTextLength())
	{
		if (m_pacSearchString && m_pacSearchString->IsBound()){
			CString strSearch;
			if (GetDlgItemText(IDC_SEARCHNAME, strSearch))
				m_pacSearchString->AddItem(strSearch, 0);
		}
		switch (methodBox.GetCurSel())
		{
		case 0:
		case 1:
			StartNewSearch();
			break;

		case 2:
		case 3:
			ShellOpenFile(CreateWebQuery());
			return;
		default:
			return;
		}
	}
}

void CSearchDlg::OnTimer(UINT nIDEvent){
	CResizableDialog::OnTimer(nIDEvent);
	if (theApp.serverconnect->IsConnected()){
		CServer* toask = theApp.serverlist->GetNextSearchServer();
		if (toask == theApp.serverlist->GetServerByAddress(theApp.serverconnect->GetCurrentServer()->GetAddress(),theApp.serverconnect->GetCurrentServer()->GetPort()))
			toask = theApp.serverlist->GetNextSearchServer();

		if (toask && theApp.serverlist->GetServerCount()-1 != servercount){
			servercount++;
			theApp.serverconnect->SendUDPPacket(searchpacket,toask,false);
			searchprogress.StepIt();
		}
		else
			OnBnClickedCancels();
	}
	else
		OnBnClickedCancels();
}

void CSearchDlg::OnBnClickedCancels(){
	canceld = true;
	if (globsearch){
		delete searchpacket;
		searchpacket = NULL;
	}
	globsearch = false;
	if (global_search_timer){
		KillTimer(global_search_timer);
		global_search_timer = 0;
		searchprogress.SetPos(0);
	}
	CWnd* pWndFocus = GetFocus();
	GetDlgItem(IDC_CANCELS)->EnableWindow(false);
	if (pWndFocus == GetDlgItem(IDC_CANCELS))
		GetDlgItem(IDC_SEARCHNAME)->SetFocus();
	GetDlgItem(IDC_STARTS)->EnableWindow(true);
}


void CSearchDlg::LocalSearchEnd(uint16 count){
	if (!canceld && count > MAX_RESULTS)
		OnBnClickedCancels();
	if (!canceld){	
		if (!globsearch){
			GetDlgItem(IDC_STARTS)->EnableWindow(true);
			CWnd* pWndFocus = GetFocus();
			GetDlgItem(IDC_CANCELS)->EnableWindow(false);
			if (pWndFocus == GetDlgItem(IDC_CANCELS))
				GetDlgItem(IDC_SEARCHNAME)->SetFocus();
		}
		else{
			VERIFY( (global_search_timer = SetTimer(1, 750, 0)) );
		}
	}
}

void CSearchDlg::AddUDPResult(uint16 count){
	if (!canceld && count > MAX_RESULTS)
		OnBnClickedCancels();
}

void CSearchDlg::OnBnClickedSdownload(){
	//start download(s)
	DownloadSelected();
}

BOOL CSearchDlg::PreTranslateMessage(MSG* pMsg) 
{
	if((pMsg->message == WM_KEYDOWN) && (pMsg->wParam == VK_ESCAPE))
	   	return FALSE;

	if( m_pacSearchString && m_pacSearchString->IsBound() && ((pMsg->message == WM_KEYDOWN) && (pMsg->wParam == VK_DELETE) && (pMsg->hwnd == GetDlgItem(IDC_SEARCHNAME)->m_hWnd) && (GetAsyncKeyState(VK_MENU)<0 || GetAsyncKeyState(VK_CONTROL)<0)) )
		m_pacSearchString->Clear();

   	if((pMsg->message == WM_KEYDOWN) && (pMsg->wParam == VK_RETURN)){
	   	if (pMsg->hwnd == GetDlgItem(IDC_SEARCHLIST)->m_hWnd)
			OnBnClickedSdownload();
	   	else if (pMsg->hwnd == GetDlgItem(IDC_SEARCHNAME)->m_hWnd || 
				 pMsg->hwnd == GetDlgItem(IDC_TypeSearch)->m_hWnd || 
				 pMsg->hwnd == GetDlgItem(IDC_EDITSEARCHMIN)->m_hWnd || 
				 pMsg->hwnd == GetDlgItem(IDC_EDITSEARCHMAX)->m_hWnd || 
				 pMsg->hwnd == GetDlgItem(IDC_EDITSEARCHAVAIBILITY)->m_hWnd || 
				 pMsg->hwnd == GetDlgItem(IDC_COMBO1)->m_hWnd || 
				 pMsg->hwnd == GetDlgItem(IDC_EDITSEARCHEXTENSION)->m_hWnd){
			if (m_pacSearchString && m_pacSearchString->IsBound() && pMsg->hwnd == GetDlgItem(IDC_SEARCHNAME)->m_hWnd){
				CString strText;
				GetDlgItem(IDC_SEARCHNAME)->GetWindowText(strText);
				if (!strText.IsEmpty()){
					GetDlgItem(IDC_SEARCHNAME)->SetWindowText(_T("")); // this seems to be the only chance to let the dropdown list to disapear
					GetDlgItem(IDC_SEARCHNAME)->SetWindowText(strText);
					((CEdit*)GetDlgItem(IDC_SEARCHNAME))->SetSel(strText.GetLength(), strText.GetLength());
				}
			}
			OnBnClickedStarts();
   		}
	}

	return CResizableDialog::PreTranslateMessage(pMsg);
}

void CSearchDlg::OnNMDblclkSearchlist(NMHDR *pNMHDR, LRESULT *pResult){
	OnBnClickedSdownload();
	*pResult = 0;
}

ULONG GetSearchSize(const CString& strExpr)
{
	ULONG ulNum;
	TCHAR szUnit[40];
	int iArgs = _stscanf(strExpr, _T("%u%s"), &ulNum, szUnit);
	if (iArgs <= 0)
		return 0;
	if (iArgs == 2){
		CString strUnits(szUnit);
		strUnits.Trim();
		if (!strUnits.IsEmpty()){
			if (strUnits.CompareNoCase(_T("b")) == 0 || strUnits.CompareNoCase(_T("byte")) == 0 || strUnits.CompareNoCase(_T("bytes")) == 0)
				return ulNum * 1U; // Bytes
			else if (strUnits.CompareNoCase(_T("k")) == 0 || strUnits.CompareNoCase(_T("kb")) == 0 || strUnits.CompareNoCase(_T("kbyte")) == 0 || strUnits.CompareNoCase(_T("kbytes")) == 0)
				return ulNum * 1024U; // KBytes
			else if (strUnits.CompareNoCase(_T("m")) == 0 || strUnits.CompareNoCase(_T("mb")) == 0 || strUnits.CompareNoCase(_T("mbyte")) == 0 || strUnits.CompareNoCase(_T("mbytes")) == 0)
				return ulNum * 1024U*1024; // MBytes
			else if (strUnits.CompareNoCase(_T("g")) == 0 || strUnits.CompareNoCase(_T("gb")) == 0 || strUnits.CompareNoCase(_T("gbyte")) == 0 || strUnits.CompareNoCase(_T("gbytes")) == 0)
				return ulNum * 1024U*1024U*1024U; // GBytes
			else{
				AfxMessageBox(GetResString(IDS_SEARCH_EXPRERROR) + _T("\n\n") + GetResString(IDS_SEARCH_INVALIDMINMAX));
				return -1;
			}
		}
	}

	return ulNum * 1024U*1024U; // Default = MBytes
}

CString	CSearchDlg::CreateWebQuery()
{
	CString query,tosearch;
	unsigned long num;
	switch (methodBox.GetCurSel())
	{
	case 2:
		query = "http://www.filedonkey.com/search.html?";
		GetDlgItem(IDC_SEARCHNAME)->GetWindowText(tosearch);
		query += "pattern="+ToQueryString(tosearch);
		GetDlgItem(IDC_TypeSearch)->GetWindowText(tosearch);
		if (GetResString(IDS_SEARCH_AUDIO)==tosearch)
			tosearch="Audio";
		else if (GetResString(IDS_SEARCH_VIDEO)==tosearch)
			tosearch="Video";
		else if (GetResString(IDS_SEARCH_PRG)==tosearch)
			tosearch="Pro";
		else
			tosearch="All";
		query +="&media=" + tosearch + "&requestby=emule";

		GetDlgItem(IDC_EDITSEARCHMIN)->GetWindowText(tosearch);
		if ((num=GetSearchSize(tosearch)) == (ULONG)-1)
			return "";
		if (num>0)
			query.AppendFormat("&min_size=%u",num);
		
		GetDlgItem(IDC_EDITSEARCHMAX)->GetWindowText(tosearch);
		if ((num=GetSearchSize(tosearch)) == (ULONG)-1)
			return "";
		if (num>0)
			query.AppendFormat("&max_size=%u",num);

		break;
	case 3:
		// www.jigle.com
		//
		// Type				ID
		// --------------------------
		//   Any			0
		//   Audio			1
		//   Video			2
		//   Image			3
		//   Program		4
		//   Document		5
		//   Collection		6

		query = "http://www.jigle.com/search?";
		GetDlgItem(IDC_SEARCHNAME)->GetWindowText(tosearch);
		query += "p="+ToQueryString(tosearch) + "&ma=";
		GetDlgItem(IDC_EDITSEARCHAVAIBILITY)->GetWindowText(tosearch);
		if (atol(tosearch)>1000) tosearch="1000";
		else if (atoi(tosearch)>500) tosearch="500";
		else if (atoi(tosearch)>200) tosearch="200";
		else if (atoi(tosearch)>100) tosearch="100";
		else if (atoi(tosearch)>50) tosearch="50";
		else if (atoi(tosearch)>20) tosearch="20";
		else if (atoi(tosearch)>10) tosearch="10";
		else tosearch="1";
		query += tosearch +"&v=0&d=1&a=0&l=10";		//Cax2 might want to change to l=25 for more results per page... 
		GetDlgItem(IDC_TypeSearch)->GetWindowText(tosearch);
		if (GetResString(IDS_SEARCH_AUDIO)==tosearch)
			tosearch="1";
		else if (GetResString(IDS_SEARCH_VIDEO)==tosearch)
			tosearch="2";
		else if (GetResString(IDS_SEARCH_PICS)==tosearch)
			tosearch="3";
		else if (GetResString(IDS_SEARCH_PRG)==tosearch)
			tosearch="4";
		else
			tosearch="0";
		query +="&t=" +tosearch;
		GetDlgItem(IDC_EDITSEARCHEXTENSION)->GetWindowText(tosearch);
		query +="&x=" + ToQueryString(tosearch);
		GetDlgItem(IDC_EDITSEARCHMIN)->GetWindowText(tosearch);
		if ((num=GetSearchSize(tosearch)) == (ULONG)-1)
			return "";
		tosearch.Format("%u",((num>0)?num:1));
		query +="&sl=" +tosearch;
		GetDlgItem(IDC_EDITSEARCHMAX)->GetWindowText(tosearch);
		if ((num=GetSearchSize(tosearch)) == (ULONG)-1)
			return "";
		tosearch.Format(((num>0)?"%u":""),num);
		query +="&su=" + tosearch;
		break;
	default:
		return "";
	}
	return query;
}


void CSearchDlg::DownloadSelected() {
	int index = -1; 
	POSITION pos = searchlistctrl.GetFirstSelectedItemPosition(); 
	while(pos != NULL) 
	{ 
		index = searchlistctrl.GetNextSelectedItem(pos); 
		if(index > -1) {
			theApp.downloadqueue->AddSearchToDownload((CSearchFile*)searchlistctrl.GetItemData(index ),m_cattabs.GetCurSel());
			searchlistctrl.Update(index);
		}
	} 
}


void CSearchDlg::Localize(){
	searchlistctrl.Localize();

	m_ctrlSearchFrm.SetWindowText(GetResString(IDS_SW_SEARCHBOX));
	m_ctrlDirectDlFrm.SetWindowText(GetResString(IDS_SW_DIRECTDOWNLOAD));

	GetDlgItem(IDC_MSTATIC3)->SetWindowText(GetResString(IDS_SW_NAME));
	GetDlgItem(IDC_MSTATIC7)->SetWindowText(GetResString(IDS_SW_TYPE));
	GetDlgItem(IDC_STATIC_FILTER)->SetWindowText(GetResString(IDS_FILTER));
	GetDlgItem(IIDC_FSTATIC2)->SetWindowText(GetResString(IDS_SW_LINK));
	GetDlgItem(IDC_STARTS)->SetWindowText(GetResString(IDS_SW_START));
	GetDlgItem(IDC_CANCELS)->SetWindowText(GetResString(IDS_CANCEL));
	GetDlgItem(IDC_CLEARALL)->SetWindowText(GetResString(IDS_REMOVEALLSEARCH));
	GetDlgItem(IDC_RESULTS_LBL)->SetWindowText(GetResString(IDS_SW_RESULT));
	GetDlgItem(IDC_SDOWNLOAD)->SetWindowText(GetResString(IDS_SW_DOWNLOAD));
	GetDlgItem(IDC_MSTATIC7)->SetWindowText(GetResString(IDS_TYPE));
	GetDlgItem(IDC_SEARCH_RESET)->SetWindowText(GetResString(IDS_PW_RESET));	
	GetDlgItem(IDC_SEARCHMINSIZE)->SetWindowText(GetResString(IDS_SEARCHMINSIZE)+":");
	GetDlgItem(IDC_SEARCHMAXSIZE)->SetWindowText(GetResString(IDS_SEARCHMAXSIZE)+":");
	GetDlgItem(IDC_SEARCHEXTENTION)->SetWindowText(GetResString(IDS_SEARCHEXTENTION)+":");
	GetDlgItem(IDC_SEARCHAVAIL)->SetWindowText(GetResString(IDS_SEARCHAVAIL)+":");
	GetDlgItem(IDC_METH)->SetWindowText(GetResString(IDS_METHOD));
	GetDlgItem(IDC_SD_MB1)->SetWindowText(GetResString(IDS_MBYTES));
	GetDlgItem(IDC_SD_MB2)->SetWindowText(GetResString(IDS_MBYTES));

	while (methodBox.GetCount()>0) methodBox.DeleteString(0);
	methodBox.AddString(GetResString(IDS_SERVER));
	methodBox.AddString(GetResString(IDS_GLOBALSEARCH));
	methodBox.AddString("Filedonkey");
	methodBox.AddString("Jigle");
	methodBox.SetCurSel(0);

	while (Stypebox.GetCount()>0) Stypebox.DeleteString(0);
	Stypebox.AddString(GetResString(IDS_SEARCH_ANY));
	Stypebox.AddString(GetResString(IDS_SEARCH_ARC));
	Stypebox.AddString(GetResString(IDS_SEARCH_AUDIO));
	Stypebox.AddString(GetResString(IDS_SEARCH_CDIMG));
	Stypebox.AddString(GetResString(IDS_SEARCH_PICS));
	Stypebox.AddString(GetResString(IDS_SEARCH_PRG));
	Stypebox.AddString(GetResString(IDS_SEARCH_VIDEO));
	Stypebox.SetCurSel(Stypebox.FindString(-1,GetResString(IDS_SEARCH_ANY)));
}

void CSearchDlg::OnBnClickedClearall()
{
	OnBnClickedCancels();
	DeleteAllSearchs();
}

static CSearchExpr _SearchExpr;
CStringArray _astrParserErrors;

#ifdef _DEBUG
static char _chLastChar = 0;

bool DumpSearchTree(int& iExpr, const CSearchExpr& rSearchExpr)
{
	if (iExpr >= rSearchExpr.m_aExpr.GetCount())
		return false;
	CString strTok = rSearchExpr.m_aExpr[iExpr++];
	if (strTok == SEARCHOPTOK_AND || strTok == SEARCHOPTOK_OR || strTok == SEARCHOPTOK_NOT){
		if (_chLastChar != '(' && _chLastChar != '\0')
			TRACE(" ");
		TRACE("(%s ", strTok.Mid(1));
		_chLastChar = '(';
		DumpSearchTree(iExpr, rSearchExpr);
		DumpSearchTree(iExpr, rSearchExpr);
		TRACE(")");
		_chLastChar = ')';
	}
	else{
		if (_chLastChar != '(' && _chLastChar != '\0')
			TRACE(" ");
		TRACE("\"%s\"", strTok);
		_chLastChar = '\1';
	}
	return true;
}

bool DumpSearchTree(const CSearchExpr& rSearchExpr)
{
	_chLastChar = '\0';
	int iExpr = 0;
	return DumpSearchTree(iExpr, rSearchExpr);
}
#endif//!_DEBUG

void ParsedSearchExpression(const CSearchExpr* pexpr)
{
	int iOpAnd = 0;
	int iOpOr = 0;
	int iOpNot = 0;
	TRACE("Expr:  ");
	for (int i = 0; i < pexpr->m_aExpr.GetCount(); i++){
		CString str(pexpr->m_aExpr[i]);
		if (str == SEARCHOPTOK_AND){
			iOpAnd++;
			TRACE("%s ", str.Mid(1));
		}
		else if (str == SEARCHOPTOK_OR){
			iOpOr++;
			TRACE("%s ", str.Mid(1));
		}
		else if (str == SEARCHOPTOK_NOT){
			iOpNot++;
			TRACE("%s ", str.Mid(1));
		}
		else
			TRACE("\"%s\" ", str);
	}
	TRACE("\n");

	if (iOpAnd + iOpOr + iOpNot > 10)
		yyerror(GetResString(IDS_SEARCH_TOOCOMPLEX));

	_SearchExpr.m_aExpr.RemoveAll();
	// optimize search expression, if no OR nor NOT specified
	if (iOpAnd > 0 && iOpOr == 0 && iOpNot == 0){
		CString strAndTerms;
		for (int i = 0; i < pexpr->m_aExpr.GetCount(); i++){
			if (pexpr->m_aExpr[i] != SEARCHOPTOK_AND){
				if (!strAndTerms.IsEmpty())
					strAndTerms += _T(' ');
				strAndTerms += pexpr->m_aExpr[i];
			}
		}
		ASSERT( _SearchExpr.m_aExpr.GetCount() == 0);
		_SearchExpr.m_aExpr.Add(strAndTerms);
	}
	else
		_SearchExpr.m_aExpr.Append(pexpr->m_aExpr);

#ifdef _DEBUG
	TRACE("Tree:  ");
	DumpSearchTree(_SearchExpr);
#endif
	TRACE("\n");
}

bool GetSearchPacket(CSafeMemFile& data,
					 const CString& strSearchString, const CString& strLocalizedType,
					 ULONG ulMinSize, ULONG ulMaxSize, int iAvailability, 
					 const CString& strExtension, bool bAllowEmptySearchString)
{
	static const uchar _aucSearchType[3]         = {       0x01, 0x00, FT_FILETYPE	 };
	static const uchar _aucSearchExtension[3]    = {       0x01, 0x00, FT_FILEFORMAT };
	static const uchar _aucSearchAvailability[4] = { 0x01, 0x01, 0x00, FT_SOURCES    };
	static const uchar _aucSearchMin[4]			 = { 0x01, 0x01, 0x00, FT_FILESIZE   };
	static const uchar _aucSearchMax[4]			 = { 0x02, 0x01, 0x00, FT_FILESIZE   };

	static const byte stringParameter  = 1;
	static const byte typeParameter    = 2;
	static const byte numericParameter = 3;

	static const uint16 andParameter = 0x0000;	// 0x00, 0x00
	static const uint16 orParameter  = 0x0100;	// 0x00, 0x01
	static const uint16 notParameter = 0x0200;	// 0x00, 0x02

	//TODO: use 'GetE2DKFileTypeSearchTerm'
	CString strType;
	if (GetResString(IDS_SEARCH_AUDIO) == strLocalizedType)
		strType = _T("Audio");
	else if (GetResString(IDS_SEARCH_VIDEO) == strLocalizedType)
		strType = _T("Video");
	else if (GetResString(IDS_SEARCH_PRG) == strLocalizedType)
		strType = _T("Pro");
	else if (GetResString(IDS_SEARCH_PICS) == strLocalizedType)
		strType = _T("Image");
	else if (GetResString(IDS_SEARCH_ARC) == strLocalizedType){
		// eDonkeyHybrid 0.48 uses type "Pro" for archives files
		// www.filedonkey.com uses type "Pro" for archives files
		//strType = _T("Pro"); //TODO: Use a proper type - filtering the search results locally is evil!
	}
	else if (GetResString(IDS_SEARCH_CDIMG) == strLocalizedType){
		// eDonkeyHybrid 0.48 uses *no* type for iso/nrg/cue/img files
		// www.filedonkey.com uses type "Pro" for CD-image files
		//strType = _T("Pro"); //TODO: Use a proper type - filtering the search results locally is evil!
	}
	else{
		//TODO: Support "Doc" types
		ASSERT( GetResString(IDS_SEARCH_ANY) == strLocalizedType );
	}

	if (!bAllowEmptySearchString && strSearchString.IsEmpty() && strExtension.IsEmpty())
		return false;

	_astrParserErrors.RemoveAll();
	_SearchExpr.m_aExpr.RemoveAll();
	if (!strSearchString.IsEmpty()){
	    LexInit(strSearchString);
	    int iParseResult = yyparse();
	    LexFree();
	    if (_astrParserErrors.GetSize() > 0){
		    _SearchExpr.m_aExpr.RemoveAll();
		    AfxMessageBox(GetResString(IDS_SEARCH_EXPRERROR) + _T("\n\n") + _astrParserErrors[_astrParserErrors.GetSize() - 1], MB_ICONWARNING);
		    return false;
	    }
	    else if (iParseResult != 0){
		    _SearchExpr.m_aExpr.RemoveAll();
		    AfxMessageBox(GetResString(IDS_SEARCH_EXPRERROR) + _T("\n\n") + GetResString(IDS_SEARCH_GENERALERROR), MB_ICONWARNING);
		    return false;
	    }
	}

	// get total nr. of search terms
	int iTotalTerms = 0;
	if (!strExtension.IsEmpty())
		iTotalTerms++;
	// Because the availability is used with the 'Min' operator, we decrement the entered value by one to get
	// more 'useable' results from the server (from a user's POV).
	if (iAvailability >= 0)
		iTotalTerms++;
	if (ulMaxSize > 0)
		iTotalTerms++;
	if (ulMinSize > 0)
		iTotalTerms++;
	if (!strType.IsEmpty())
		iTotalTerms++;
	iTotalTerms += _SearchExpr.m_aExpr.GetCount();

	// create ed2k search expression
	TRACE("Data:  ");
	int iParameterCount = 0;
	if (_SearchExpr.m_aExpr.GetCount() <= 1)
	{
		// If we don't have a NOT or OR operator we use a series of AND terms which can be processed by
		// the servers with less load (request lugdunummaster)
		//
		// input:      "a" AND min=1 AND max=2
		// instead of: AND AND "a" min=1 max=2
		// we use:     AND "a" AND min=1 max=2

		if (_SearchExpr.m_aExpr.GetCount() > 0){
			if (++iParameterCount < iTotalTerms){
				data.Write(&andParameter, 2);
				TRACE("AND ");
			}
			data.Write(&stringParameter, 1);
			uint16 nSize = _SearchExpr.m_aExpr[0].GetLength();
			data.Write(&nSize, 2);
			data.Write((LPCSTR)_SearchExpr.m_aExpr[0], nSize);
			TRACE("\"%s\" ", _SearchExpr.m_aExpr[0]);
		}

		if (!strType.IsEmpty()){
			if (++iParameterCount < iTotalTerms){
				data.Write(&andParameter, 2);
				TRACE("AND ");
			}
			data.Write(&typeParameter, 1);
			uint16 nSize = strType.GetLength();
			data.Write(&nSize, 2);
			data.Write((LPCSTR)strType, nSize);
			data.Write(_aucSearchType, sizeof _aucSearchType);
			TRACE("type=\"%s\" ", strType);
		}
		
		if (ulMinSize > 0){
			if (++iParameterCount < iTotalTerms){
				data.Write(&andParameter, 2);
				TRACE("AND ");
			}
			data.Write(&numericParameter, 1);
			data.Write(&ulMinSize, 4);
			data.Write(_aucSearchMin, sizeof _aucSearchMin);
			TRACE("min=%u ", ulMinSize);
		}

		if (ulMaxSize > 0){
			if (++iParameterCount < iTotalTerms){
				data.Write(&andParameter, 2);
				TRACE("AND ");
			}
			data.Write(&numericParameter, 1);
			data.Write(&ulMaxSize, 4);
			data.Write(_aucSearchMax, sizeof _aucSearchMax);
			TRACE("max=%u ", ulMaxSize);
		}
		
		if (iAvailability >= 0){
			if (++iParameterCount < iTotalTerms){
				data.Write(&andParameter, 2);
				TRACE("AND ");
			}
			data.Write(&numericParameter, 1);
			data.Write(&iAvailability, 4);
			data.Write(_aucSearchAvailability, sizeof _aucSearchAvailability);
			TRACE("avail=%u ", iAvailability);
		}

		if (!strExtension.IsEmpty()){
			if (++iParameterCount < iTotalTerms){
		    	data.Write(&andParameter,2);
				TRACE("AND ");
			}
			CString strData(strExtension);
			data.Write(&typeParameter, 1);
			uint16 nSize = strData.GetLength();
			data.Write(&nSize, 2);
			data.Write((LPCSTR)strData, nSize);
			data.Write(_aucSearchExtension, sizeof _aucSearchExtension);
			TRACE("ext=\"%s\" ", strData);
		}

		ASSERT( iParameterCount == iTotalTerms );
	}
	else
	{
		if (!strExtension.IsEmpty()){
			if (++iParameterCount < iTotalTerms){
				data.Write(&andParameter, 2);
				TRACE("AND ");
			}
		}

		if (iAvailability >= 0){
			if (++iParameterCount < iTotalTerms){
				data.Write(&andParameter,2);
				TRACE("AND ");
			}
		}
	  
		if (ulMaxSize > 0){
			if (++iParameterCount < iTotalTerms){
				data.Write(&andParameter,2);
				TRACE("AND ");
			}
		}
        
		if (ulMinSize > 0){
			if (++iParameterCount < iTotalTerms){
				data.Write(&andParameter,2);
				TRACE("AND ");
			}
		}
        
		if (!strType.IsEmpty()){
			if (++iParameterCount < iTotalTerms){
				data.Write(&andParameter,2);
				TRACE("AND ");
			}
		}
        
		for (int j = 0; j < _SearchExpr.m_aExpr.GetCount(); j++){
			CString str(_SearchExpr.m_aExpr[j]);
			if (str == SEARCHOPTOK_AND){
				data.Write(&andParameter,2);
				TRACE("AND ");
			}
			else if (str == SEARCHOPTOK_OR){
				data.Write(&orParameter, 2);
				TRACE("OR ");
			}
			else if (str == SEARCHOPTOK_NOT){
				data.Write(&notParameter, 2);
				TRACE("NOT ");
			}
			else{
				data.Write(&stringParameter, 1);
				uint16 nSize = str.GetLength();
				data.Write(&nSize, 2);
				data.Write((LPCSTR)str, nSize);
				TRACE("\"%s\" ", str);
			}
		}

		if (!strType.IsEmpty()){
			data.Write(&typeParameter, 1);
			uint16 nSize = strType.GetLength();
			data.Write(&nSize, 2);
			data.Write((LPCSTR)strType, nSize);
			data.Write(_aucSearchType, sizeof _aucSearchType);
			TRACE("type=\"%s\" ", strType);
		}

		if (ulMinSize > 0){
			data.Write(&numericParameter, 1);
			data.Write(&ulMinSize, 4);
			data.Write(_aucSearchMin, sizeof _aucSearchMin);
			TRACE("min=%u ", ulMinSize);
		}

		if (ulMaxSize > 0){
			data.Write(&numericParameter, 1);
			data.Write(&ulMaxSize, 4);
			data.Write(_aucSearchMax, sizeof _aucSearchMax);
			TRACE("max=%u ", ulMaxSize);
		}

		if (iAvailability >= 0){
			data.Write(&numericParameter, 1);
			data.Write(&iAvailability, 4);
			data.Write(_aucSearchAvailability, sizeof _aucSearchAvailability);
			TRACE("avail=%u ", iAvailability);
		}

		if (!strExtension.IsEmpty()){
			CString strData(strExtension);
			data.Write(&typeParameter, 1);
			uint16 nSize = strData.GetLength();
			data.Write(&nSize, 2);
			data.Write((LPCSTR)strData, nSize);
			data.Write(_aucSearchExtension, sizeof _aucSearchExtension);
			TRACE("ext=\"%s\" ", strData);
		}
	}

	TRACE("\n\n");
	_SearchExpr.m_aExpr.RemoveAll();
	return true;
}

void CSearchDlg::DoNewSearch(const CString& strSearchString, const CString& strLocalizedType,
							 ULONG ulMinSize, ULONG ulMaxSize, int iAvailability, 
							 const CString& strExtension, bool doGlobal){

	if (!theApp.serverconnect->IsConnected())
		return;

	CSafeMemFile data(100);
	if (!GetSearchPacket(data, strSearchString, strLocalizedType, ulMinSize, ulMaxSize, iAvailability, strExtension, false) || data.GetLength() == 0)
		return;

	m_nSearchID++;
	theApp.searchlist->NewSearch(&searchlistctrl, strLocalizedType, m_nSearchID);
	canceld = false;

	Packet* packet = new Packet(&data);
	packet->opcode = OP_SEARCHREQUEST;
	theApp.uploadqueue->AddUpDataOverheadServer(packet->size);
	theApp.serverconnect->SendPacket(packet,false);

	if (doGlobal && theApp.serverconnect->IsUDPSocketAvailable()){
		if( theApp.glob_prefs->Score() ){
			theApp.serverlist->ResetSearchServerPos();
		}

		if (globsearch){
			delete searchpacket;
			searchpacket = NULL;
		}
		searchpacket = packet;
		searchpacket->opcode = OP_GLOBSEARCHREQ;
		servercount = 0;
		searchprogress.SetRange32(0,theApp.serverlist->GetServerCount()-1);
		globsearch = true;
	}
	else{
		globsearch = false;
		delete packet;
	}
	CreateNewTab(strSearchString, m_nSearchID);
}
	
void CSearchDlg::StartNewSearch(){
	if (!theApp.serverconnect->IsConnected())
		AfxMessageBox(GetResString(IDS_ERR_NOTCONNECTED),MB_ICONASTERISK);
	else{
		CString strSearchString;
		GetDlgItem(IDC_SEARCHNAME)->GetWindowText(strSearchString);
		strSearchString.Trim();

		CString strSearchType;
		GetDlgItem(IDC_TypeSearch)->GetWindowText(strSearchType);
		//strSearchType.Trim();

		CString strMinSize;
		GetDlgItem(IDC_EDITSEARCHMIN)->GetWindowText(strMinSize);
		ULONG ulMinSize = GetSearchSize(strMinSize);
		if (ulMinSize == (ULONG)-1)
			return;

		CString strMaxSize;
		GetDlgItem(IDC_EDITSEARCHMAX)->GetWindowText(strMaxSize);
		ULONG ulMaxSize = GetSearchSize(strMaxSize);
		if (ulMaxSize == (ULONG)-1)
			return;
		
		// 06-Mai-2003: always sending the max-size is no longer needed
		// there is no need to always send a default max size (confirmed by lugdunummaster)
		if (ulMaxSize < ulMinSize){
			ulMaxSize = 0; // TODO: Create a message box for that
			SetDlgItemText(IDC_EDITSEARCHMAX, _T(""));
		}

		CString strExtension;
		GetDlgItem(IDC_EDITSEARCHEXTENSION)->GetWindowText(strExtension);
		strExtension.Trim();
		if (!strExtension.IsEmpty() && strExtension[0] == _T('.')){
			strExtension = strExtension.Mid(1);
			SetDlgItemText(IDC_EDITSEARCHEXTENSION, strExtension);
		}

		// Because the availability is used with the 'Min' operator, we decrement the entered value by one to get
		// more 'useable' results from the server (from a user's POV). Therefore its correct to send an
		// 'Availability' of 0, because it will return all files with an 'Availability' of >= 1.
		CString strAvailability;
		GetDlgItem(IDC_EDITSEARCHAVAIBILITY)->GetWindowText(strAvailability);
		int iAvailability;
		if (strAvailability.IsEmpty())
			iAvailability = -1; // No Availability specified
		else
			iAvailability = _tstol(strAvailability) - 1;

		CWnd* pWndFocus = GetFocus();
		GetDlgItem(IDC_STARTS)->EnableWindow(false);
		if (pWndFocus == GetDlgItem(IDC_STARTS))
			GetDlgItem(IDC_SEARCHNAME)->SetFocus();
		GetDlgItem(IDC_CANCELS)->EnableWindow(true);

		DoNewSearch(strSearchString, strSearchType, ulMinSize, ulMaxSize, 
			        iAvailability, strExtension, (methodBox.GetCurSel() == 1));
	}
}

bool CSearchDlg::CreateNewTab(CString searchString,uint32 nSearchID){
    int iTabItems = searchselect.GetItemCount();
    for (int i = 0; i < iTabItems; i++){
        TCITEM tci;
        tci.mask = TCIF_PARAM;
        if (searchselect.GetItem(i, &tci) && tci.lParam == nSearchID)
			return false;
    }
	// add new tab
	TCITEM newitem;
	if (searchString.GetLength() == 0)
		searchString = "-";
	newitem.mask = TCIF_PARAM|TCIF_TEXT|TCIF_IMAGE;
	newitem.lParam = (LPARAM)nSearchID;
	newitem.pszText = searchString.GetBuffer();
	newitem.cchTextMax = (int)searchString.GetLength()+10;
	newitem.iImage = 0;
	uint16 itemnr = searchselect.InsertItem(searchselect.GetItemCount(),&newitem);
	if (!searchselect.IsWindowVisible())
		ShowSearchSelector(true);
	searchselect.SetCurSel(itemnr);
	searchlistctrl.ShowResults(nSearchID);
	return true;
}

void CSearchDlg::DeleteSearch(uint32 nSearchID){
	TCITEM item;
	item.mask = TCIF_PARAM;
	item.lParam = -1;
	int i;
	for (i = 0; i != searchselect.GetItemCount();i++){
		searchselect.GetItem(i,&item);
		if (item.lParam == nSearchID)
			break;
	}
	if (item.lParam != nSearchID)
		return;
	if ((!canceld) && (nSearchID == m_nSearchID))
		OnBnClickedCancels();  //Sony - cancel active search before closing active tab
	theApp.searchlist->RemoveResults(nSearchID);
	searchselect.DeleteItem(i);
	if (searchselect.GetItemCount()){
		searchselect.SetCurSel(0);
		item.mask = TCIF_PARAM;
		searchselect.GetItem(0,&item);
		searchlistctrl.ShowResults(item.lParam);
	}
	else{
		searchlistctrl.DeleteAllItems();
		ShowSearchSelector(false);
		searchlistctrl.NoTabs();
	}
}

void CSearchDlg::DeleteAllSearchs(){
	theApp.searchlist->Clear();
	searchlistctrl.DeleteAllItems();
	ShowSearchSelector(false);
	searchselect.DeleteAllItems();
}

void CSearchDlg::OnTcnSelchangeTab1(NMHDR *pNMHDR, LRESULT *pResult)
{
	TCITEM item;
	item.mask = TCIF_PARAM;
	int cur_sel = searchselect.GetCurSel();
	if (cur_sel == (-1))
		return;
	searchselect.GetItem(cur_sel,&item);
	searchlistctrl.ShowResults(item.lParam);
	*pResult = 0;
}


LRESULT CSearchDlg::OnCloseTab(WPARAM wparam, LPARAM lparam) {
	TCITEM item;
	item.mask = TCIF_PARAM;
	int i = (int)wparam;
	searchselect.GetItem(i,&item);
	int nSearchID = item.lParam;

	if ((!canceld) && (nSearchID == m_nSearchID))
		OnBnClickedCancels();  //Sony - cancel active search before closing active tab
	DeleteSearch(nSearchID);
	//theApp.searchlist->RemoveResults(nSearchID);

	return true;

}

void CSearchDlg::OnEnChangeSearchname()
{
	GetDlgItem(IDC_STARTS)->EnableWindow(true);
}

void CSearchDlg::OnBnClickedSearchReset()
{
	GetDlgItem(IDC_EDITSEARCHAVAIBILITY)->SetWindowText("");	
	GetDlgItem(IDC_EDITSEARCHEXTENSION)->SetWindowText("");	
	GetDlgItem(IDC_EDITSEARCHMIN)->SetWindowText("");	
	GetDlgItem(IDC_EDITSEARCHMAX)->SetWindowText("");	
	GetDlgItem(IDC_SEARCHNAME)->SetWindowText("");
	Stypebox.SetCurSel(Stypebox.FindString(-1,GetResString(IDS_SEARCH_ANY)));
	methodBox.SetCurSel(0);

	OnEnChangeSearchname();
}

void CSearchDlg::UpdateCatTabs() {
	int oldsel=m_cattabs.GetCurSel();
	m_cattabs.DeleteAllItems();
	for (int ix=0;ix<theApp.glob_prefs->GetCatCount();ix++)
		m_cattabs.InsertItem(ix,(ix==0)?GetResString(IDS_ALL):theApp.glob_prefs->GetCategory(ix)->title);
	if (oldsel>=m_cattabs.GetItemCount() || oldsel==-1)
		oldsel=0;

	m_cattabs.SetCurSel(oldsel);
	int flag;
	flag=(m_cattabs.GetItemCount()>1) ? SW_SHOW:SW_HIDE;
	
	this->GetDlgItem(IDC_CATTAB2)->ShowWindow(flag);
	this->GetDlgItem(IDC_STATIC_DLTOof)->ShowWindow(flag);
}

void CSearchDlg::SearchClipBoard() {
	if (m_guardCBPrompt) return;

	CString clpbrd=theApp.CopyTextFromClipboard();
	if (clpbrd=="") return;

	if (clpbrd.Compare(m_lastclpbrd)==0) return;

	if (clpbrd.Left(13).CompareNoCase("ed2k://|file|")==0) {
		m_guardCBPrompt=true;
		if (AfxMessageBox(GetResString(IDS_ED2KLINKFIX) + _T("\r\n\r\n") + GetResString(IDS_ADDDOWNLOADSFROMCB)+_T("\r\n") + clpbrd,MB_YESNO|MB_SETFOREGROUND)==IDYES)
			AddEd2kLinksToDownload(clpbrd,0);
	}
	m_lastclpbrd=clpbrd;
	m_guardCBPrompt=false;

}

void CSearchDlg::AddEd2kLinksToDownload(CString strlink,uint8 cat){
	CString resToken;
	int curPos=0;

	resToken= strlink.Tokenize("\t\n\r",curPos);
	while (resToken != "")
	{
		
		if (resToken.Right(1)!="/") resToken+="/";
		try {
			CED2KLink* pLink = CED2KLink::CreateLinkFromUrl(resToken.Trim());
			_ASSERT( pLink !=0 );
			if ( pLink->GetKind() == CED2KLink::kFile) {
				theApp.downloadqueue->AddFileLinkToDownload(pLink->GetFileLink(),cat);
			} else {
				delete pLink; // [i_a] memleak
				throw CString(_T("bad link"));
			}
			delete pLink;
		} catch(CString error){
			OUTPUT_DEBUG_TRACE();
			char buffer[200];
			sprintf(buffer,GetResString(IDS_ERR_INVALIDLINK),error.GetBuffer());
			theApp.emuledlg->AddLogLine(true, GetResString(IDS_ERR_LINKERROR), buffer);
		}
		resToken= strlink.Tokenize("\t\n\r",curPos);
	}
}

void CSearchDlg::OnEnKillfocusElink()
{
	CString content;
	GetDlgItem(IDC_ELINK)->GetWindowText(content);
	if (content.GetLength()==0 || content.Find('\n')==-1) return;
	content.Replace("\n","\r\n");
	content.Replace("\r\r","\r");
	GetDlgItem(IDC_ELINK)->SetWindowText(content);
}
CString	CSearchDlg::ToQueryString(CString str){
	CString sTmp = URLEncode(str);
	sTmp.Replace("%20","+");
	return sTmp;
}

void CSearchDlg::ShowSearchSelector(bool visible){
	WINDOWPLACEMENT wpSearchWinPos;
	WINDOWPLACEMENT wpSelectWinPos;
	searchselect.GetWindowPlacement(&wpSelectWinPos);
	searchlistctrl.GetWindowPlacement(&wpSearchWinPos);
	if (visible)
		wpSearchWinPos.rcNormalPosition.top = wpSelectWinPos.rcNormalPosition.bottom;
	else
		wpSearchWinPos.rcNormalPosition.top = wpSelectWinPos.rcNormalPosition.top;
	searchselect.ShowWindow( ((visible)? SW_SHOW : SW_HIDE) );
	RemoveAnchor(searchlistctrl);
	searchlistctrl.SetWindowPlacement(&wpSearchWinPos);
	AddAnchor(searchlistctrl,TOP_LEFT,BOTTOM_RIGHT);
	GetDlgItem(IDC_CLEARALL)->ShowWindow((visible) ? SW_SHOW:SW_HIDE);
}

BOOL CSearchDlg::SaveSearchStrings()
{
	if (m_pacSearchString == NULL)
		return FALSE;
	return m_pacSearchString->SaveList(CString(theApp.glob_prefs->GetConfigDir()) + _T("\\") SEARCH_STRINGS_PROFILE);
}
